#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/highmem.h>
#include <linux/pagemap.h>
#include <asm/arch/spi.h>
#ifdef CONFIG_DEVFS_FS
#include <linux/devfs_fs_kernel.h>
#endif

#define DEBUG_SPIDEV

#define SPI_SLAVE_MAJOR		154	/* registered device number */


#define MOZART_SLAVE_IOC_MAGIC  		0xD0	// To Be Defined
#define MOZART_GET_FLOW_CONTROL			_IOW(MOZART_SLAVE_IOC_MAGIC, 0,long)
#define MOZART_SEND_READ_SLAVE_RDY		_IOW(MOZART_SLAVE_IOC_MAGIC, 5,long)
#define MOZART_SEND_WRITE_SLAVE_RDY		_IOW(MOZART_SLAVE_IOC_MAGIC, 6,long)

/* ......................................................................... */

/*
 * Read or Write to SPI bus.
 */
static ssize_t spislavedev_rd_wr(struct file *file, char *buf, size_t count, loff_t *offset)
{
	unsigned int spi_device = (unsigned int) file->private_data;

	struct mm_struct * mm;
	struct page ** maplist;
	struct spi_transfer_list* list;
	int    pgcount;

	unsigned int ofs, pagelen;
	int res, i, err;
	printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 

	if (!count) {
		return 0;
	}
	printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 
	
	list = kmalloc(sizeof(struct spi_transfer_list), GFP_KERNEL);
	if (!list) {
		return -ENOMEM;
	}
	printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 

	mm = current->mm;

	pgcount = ((unsigned long)buf+count+PAGE_SIZE-1)/PAGE_SIZE - (unsigned long)buf/PAGE_SIZE;

	if (pgcount >= MAX_SPI_TRANSFERS) {
		kfree(list);
		return -EFBIG;
	}
	printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 

	maplist = kmalloc (pgcount * sizeof (struct page *), GFP_KERNEL);

	if (!maplist) {
		kfree(list);
		return -ENOMEM;
	}
	printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 
	flush_cache_all();
	down_read(&mm->mmap_sem);
	err= get_user_pages(current, mm, (unsigned long)buf, pgcount, 1, 0, maplist, NULL);
	up_read(&mm->mmap_sem);

	if (err < 0) {
		kfree(list);
		kfree(maplist);
		return err;
	}
	printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 
	pgcount = err;

#ifdef DEBUG_SPIDEV
	printk("spislavedev_rd_rw: %i %i\n", count, pgcount);
#endif
	/* Set default return value = transfer length */
	res = count;

	/*
	 * At this point, the virtual area buf[0] .. buf[count-1] will have
	 * corresponding pages mapped in the physical memory and locked until
	 * we unmap the kiobuf.  The pages cannot be swapped out or moved
	 * around.
	 */
	ofs = (unsigned long) buf & (PAGE_SIZE -1);
	pagelen = PAGE_SIZE - ofs;
	if (count < pagelen)
		pagelen = count;
	printk("Offset to read %d\n",ofs);
	for (i = 0; i < pgcount; i++) {
		flush_dcache_page(maplist[i]);
		printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 

		list->tx[i] = list->rx[i] = page_address(maplist[i]) + ofs;
		list->txlen[i] = list->rxlen[i] = pagelen;

#ifdef DEBUG_SPIDEV
		printk("  %i: %x  (%i)\n", i, list->tx[i], list->txlen[i]);
#endif

		ofs = 0;	/* all subsequent transfers start at beginning of a page */
		count = count - pagelen;
		pagelen = (count < PAGE_SIZE) ? count : PAGE_SIZE;
	}
	list->nr_transfers = pgcount;

	/* Perform transfer on SPI bus */
	//printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 
	//spi_access_bus(spi_device);
	printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 
	spi_transfer(list);
	//spi_release_bus(spi_device);
	//printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 

	while (pgcount--) {
		page_cache_release (maplist[pgcount]);
	}
	flush_cache_all();

	kfree(maplist);
	kfree(list);

	return res;
}

static int spislavedev_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int spi_device = MINOR(inode->i_rdev);

	printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 
	if (spi_device >= NR_SPI_DEVICES)
		return -ENODEV;

	// TODO: This interface can be used to configure the SPI bus.
	
	switch(cmd) {
		case MOZART_GET_FLOW_CONTROL:
			// TODO
			break ;
		
		case MOZART_SEND_READ_SLAVE_RDY:
			// TODO
			break ;
		
		case MOZART_SEND_WRITE_SLAVE_RDY:
			// TODO
			break;
		
		default:
			return -ENOIOCTLCMD;
	}
	return 0;
}

/*
 * Open the SPI device
 */
static int spislavedev_open(struct inode *inode, struct file *file)
{
	unsigned int spi_device = MINOR(inode->i_rdev);

	printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 
	printk("spi_device %d - NR_SPI_DEVICES %d\n",spi_device,NR_SPI_DEVICES); 
	if (spi_device >= NR_SPI_DEVICES)
	{
		return -ENODEV;
	}
	/*
	 * 'private_data' is actually a pointer, but we overload it with the
	 * value we want to store.
	 */
	file->private_data = (void *)spi_device;
	printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 
	return 0;
}

/*
 * Close the SPI device
 */
static int spislavedev_close(struct inode *inode, struct file *file)
{
	printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 
	return 0;
}

/* ......................................................................... */

static struct file_operations spislavedev_fops = {
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.read		= spislavedev_rd_wr,
	.write		= (int (*) (struct file *file, const char *buf, size_t count, loff_t *offset))spislavedev_rd_wr,
	.ioctl		= spislavedev_ioctl,
	.open		= spislavedev_open,
	.release	= spislavedev_close,
};

/*
 * Install the SPI /dev interface driver
 */
static int __init at91_spislavedev_init(void)
{
	printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 
#ifdef CONFIG_DEVFS_FS
	int i;
#endif

	if (register_chrdev(SPI_SLAVE_MAJOR, "spi_slave", &spislavedev_fops)) {
		printk(KERN_ERR "at91_spislavedev: Unable to get major %d for SPI bus\n", SPI_SLAVE_MAJOR);
		return -EIO;
	}
	else
	{
		printk("Registered at91_spislavedev Major: %d\n",SPI_SLAVE_MAJOR);
	}

	printk(KERN_INFO "AT91 SPI driver loaded\n");

	return 0;
}

/*
 * Remove the SPI /dev interface driver
 */
static void __exit at91_spislavedev_exit(void)
{
	printk("\nIn %s : %s : %d\n",__FILE__,__FUNCTION__,__LINE__); 
#ifdef CONFIG_DEVFS_FS
	int i;
	for (i = 0; i < NR_SPI_DEVICES; i++) {
		devfs_remove("spi/%d", i);
	}

	devfs_remove("spi");
#endif

	if (unregister_chrdev(SPI_SLAVE_MAJOR, "spi")) {
		printk(KERN_ERR "at91_spislavedev: Unable to release major %d for SPI bus\n", SPI_SLAVE_MAJOR);
		return;
	}
}
